/*
 * Font stuff.
 */

#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <vgamouse.h>
#include <string.h>
#include <unistd.h>

#include <hr_defines.h>
#include <hr_buttons.h>
#include <hr_general.h>
#include <hr_bitmaps.h>
#include <hr_pbm.h>

struct hr_font{
	int flags;
	unsigned char *bitmap;
	int width;
	int height;
};


/* Creates a fixed-width font.  Font must be in 16*16 matrix, since its width
 * and height are calculated run-time. Fonts are stored
 * one at a time, in row-major order.
 * Returns NULL if an error occurs.*/
struct hr_font *hr_create_font(unsigned char *name,int flags){
	struct hr_font *font=NULL;
	struct hr_bitmap *bmap;
	int c,x,y,offset,offset2;
	
	if((bmap=hr_read_pbm(name,flags))){
		/* Malloc font struct */
		font=malloc(sizeof(struct hr_font));
		font->flags=flags;
		font->width=(bmap->width)/16;
		font->height=(bmap->height)/16;
		
		/* calloc zeroed space for font bitmap. */
		if(flags & GFX_MONO)
		  font->bitmap=calloc(1+((bmap->width)*(bmap->height))/8,1);
		else if (flags & GFX_256COLOR)
		  font->bitmap=calloc((bmap->width)*(bmap->height),1);
		
		/* Convert the picture into a font.*/
		for(c=0;c<256;c++){
			for(y=0;y<(font->height);y++){
				for(x=0;x<(font->width);x++){
					/* Calc position of (topleft pixel of the character)
					 * inside bmap.*/
					offset=((c/16)*(font->height)*(bmap->width)+
							  (c%16)*(font->width));
					/* Add Y- and X-offsets within character.*/
					offset+=y*(bmap->width)+x;
					/* Now set corresponding bit or byte in font bitmap.
					 * First calc offset (in bits or bytes).*/
					offset2=c*(font->width)*(font->height)+y*(font->width)+x;
					/* Mono font. */
					if(flags & GFX_MONO){
						if(*((bmap->data)+offset/8) & (128>>(offset%8)))
						  *((font->bitmap)+offset2/8) |= (128>>(offset2%8));
					}
					/* 256color font.*/
					else if(flags & GFX_256COLOR){
						if(*(bmap->data+offset)>0)
						  *((font->bitmap)+offset2)=*((bmap->data)+offset);
					}					
				}
			}
		}
		hr_delete_bitmap(bmap);
	}
	return font;
}


/* Writes a string. Color 0 is transparent.
 * NOTE! Also handles color fonts... I'm just too lazy to
 * change the func name :-) THIS FUNCTION IS SLOW! */
void hr_monowrite(int x,int y,int color,int bcolor,
						unsigned char *text,
						struct hr_font *font,int flags)
{
	unsigned char *p;
	int x2,y2,c;
	int pixel_offset;
	int byte_offset,bit_offset;
	if((text!=NULL)&&(font!=NULL)) {
		/* Scan whole string */
		for(c=0,p=text;*p!=0;c++,p++) {
			/* Calc the offset (in bits or bytes)
			 * of pixel (0,0) in character *p */
			pixel_offset=((*p)*(font->width)*(font->height));
			/* Now scan all the pixels in the character.*/
			for(y2=0;y2<(font->height);y2++) {
				for(x2=0;x2<(font->width);x2++) {
					/* "black & white" fonts.*/
					if(font->flags & GFX_MONO) {
						/* Now divide the pixel offset into bytes and bits.*/
						byte_offset=pixel_offset/8;
						bit_offset =pixel_offset%8;
						/* If the bit is set, draw one pixel.*/
						if(*((font->bitmap)+byte_offset) & (256>>bit_offset))
						  gl_setpixel(x+c*(font->width)+x2,y+y2,color);
						/* Otherwise the pixel is background or transparent.*/
						else if(flags & GFX_SINGLECOLBGR)
						  gl_setpixel(x+c*(font->width)+x2,y+y2,bcolor);
					}
					/* Colored fonts. */
					else if(font->flags & GFX_256COLOR) {
						/* Pixel offset is now directly the offset in bytes.
						 * Color is taken from the table, and a pixel is drawn
						 * if color is non-zero. Note that parameter "color"
						 * is now the offset to be added to color.*/
						if(*((font->bitmap)+pixel_offset)){
							/* To print color fonts,
							 * 1) the font must have GFX_256COLOR set, and
							 * 2) you must give hr_monowrite() GFX_256COLOR flag.*/
							if(flags & GFX_256COLOR)
							  gl_setpixel(x+c*(font->width)+x2,y+y2,
											  (*((font->bitmap)+pixel_offset))+color);
							else
							  gl_setpixel(x+c*(font->width)+x2,y+y2,color);
						}
						/* Background */
						else if(flags & GFX_SINGLECOLBGR)
						  gl_setpixel(x+c*(font->width)+x2,y+y2,bcolor);
					}
					/* Move pixel_offset to next pixel within the character.*/
					++pixel_offset;
				}
			}
		}
	}
}


/* Writes a character, or part of it.
 * Uses 8*8 font using bit-per-pixel font.
 * Color 0 is not written.
 * No check is made whether px1/py1/px2/py2 are within legal range! */
void hr_partial_monowrite(int x,int y,int color,int bcolor,
								  int character,
								  int px1,int py1,int px2,int py2,
								  struct hr_font *font,int flags)
{
	int x2,y2,byte_offset,bit_offset;
	int character_offset,pixel_offset;
	if(font!=NULL) {
		/* Since only one character is drawn, these multiplications
		 * need to be done only once.*/
		character_offset=(character*(font->width)*(font->height));
		/* Let's make things easier */
		x-=px1;y-=py1;
		/* Scan the bits...*/
		for(y2=py1;y2<=py2;y2++) {
			/* Pixel offset must be calced more often than in plain
			 * monowrite() because not every pixel is drawn horizontally.*/
			pixel_offset=character_offset+(y2*(font->width))+px1;
			for(x2=px1;x2<=px2;x2++) {
				/* "black & white" fonts.*/
				if(font->flags & GFX_MONO) {
					/* Now divide pixel offset into bytes and bits.*/
					byte_offset=pixel_offset/8;
					bit_offset =pixel_offset%8;
					/* If the bit is set, draw one pixel.*/
					if(*((font->bitmap)+byte_offset) & (256>>bit_offset))
					  gl_setpixel(x+x2,y+y2,color);
					else if(flags & GFX_SINGLECOLBGR)
					  gl_setpixel(x+x2,y+y2,bcolor);
				}
				/* Colored fonts. */
				else if(font->flags & GFX_256COLOR) {
					/* Pixel offset is now directly the offset in bytes.
					 * Color is taken from the table, and a pixel is drawn
					 * if color is non-zero.*/
					if(*((font->bitmap)+pixel_offset)){
						if(flags & GFX_256COLOR)
						  gl_setpixel(x+x2,y+y2,
										  *((font->bitmap)+pixel_offset)+color);
						else
						  gl_setpixel(x+x2,y+y2,color);
					}
					else if(flags & GFX_SINGLECOLBGR)
					  gl_setpixel(x+x2,y+y2,bcolor);
				}
				/* Move pixel_offset to next pixel within the character.*/
				++pixel_offset;
			}
		}
	}
}


void hr_delete_font(struct hr_font *font){
	if(font!=NULL){
		free(font->bitmap);
		free(font);
	}
}
